
#ifndef CPP_START_THREAD_POOL_H
#define CPP_START_THREAD_POOL_H

#include <list>
#include <thread>
#include <memory>
#include <atomic>
#include <functional>
#include "sync_queue.h"

class ThreadPool {

    using Task = std::function<void()>;

private:

    // 线程组
    std::list<std::shared_ptr<std::thread>> mThreadGroup;

    // 任务队列
    SyncQueue<Task> mQueue;

    // 运行状态
    std::atomic_bool mRunning;

    // 标识
    std::once_flag mFlag;

private:

    void init(int numThreads) {
        mRunning = true;
        for (int i = 0; i < numThreads; ++i) {
            mThreadGroup.push_back(std::make_shared<std::thread>(&ThreadPool::run, this));
        }
    }

    void run() {
        while (mRunning) {
            std::list<Task> tasks;
            mQueue.dequeue(tasks);

            for (auto &task: tasks) {
                if (!mRunning) {
                    return;
                }
                task();
            }
        }
    }

    void stopThreadGroup() {
        mQueue.stop();
        mRunning = false;
        for (const auto &thread: mThreadGroup) {
            if (thread) {
                thread->join();
            }
        }
        mThreadGroup.clear();
    }

public:
    ThreadPool(int numThreads = std::thread::hardware_concurrency(), int maxTask = 1000) :
            mQueue(maxTask) {
        init(numThreads);
    }

    ~ThreadPool() {
        stop();
    }

    void stop() {
        std::call_once(mFlag, [this] { stopThreadGroup(); });
    }

    void add(Task &&task) {
        mQueue.enqueue(std::forward<Task>(task));
    }

    void add(const Task &task) {
        mQueue.enqueue(task);
    }

};

#endif //CPP_START_THREAD_POOL_H
